# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2021 Florian Fischer
"""gdb python script to dump the state of the runtime"""

from pathlib import Path
import sys

import gdb

gdb_py_dir = Path(__file__).parent
sys.path.append(str(gdb_py_dir))
from worker_frame_filter import WorkerFrameFilter


INDENTATION = '  '

def setup():
    gdb.execute('set pagination off')
    gdb.execute('set print pretty')
    gdb.execute('set scheduler-locking on')


def cleanup():
    gdb.execute('set pagination on')
    gdb.execute('set print pretty off')
    gdb.execute('set scheduler-locking off')


def print_queue(queue, indentation=''):
    print(f'{indentation}WSL:')
    indentation += INDENTATION

    top = queue['top']['_M_i']
    print(f'{indentation}top: {top}')

    bottom = queue['bottom']['_M_i']
    print(f'{indentation}bottom: {bottom}')

    print(f'{indentation}work: {bottom - top}')


def print_sq(io, indentation=''):
    print(f'{indentation}sq:')
    indentation += INDENTATION
    sq = io['ring']['sq']

    print(f'{indentation}ring_ptr: {sq["ring_ptr"]}')
    entries = gdb.parse_and_eval("emper::io::IoContext::workerIo->getSqEntries()")
    print(f'{indentation}entries: {entries}')

    flags = gdb.parse_and_eval("emper::io::IoContext::workerIo->getSqFlags()")
    print(f'{indentation}flags: {flags}')

    head = gdb.parse_and_eval("emper::io::IoContext::workerIo->getSqHead()")
    print(f'{indentation}head: {head}')

    tail = gdb.parse_and_eval("emper::io::IoContext::workerIo->getSqTail()")
    print(f'{indentation}tail: {tail}')

    unsubmitted = tail - head
    print(f'{indentation}unsubmitted: {unsubmitted}')

    # print(sq)


def print_cq(io, indentation=''):
    print(f'{indentation}cq:')
    indentation += INDENTATION
    cq = io['ring']['cq']

    print(f'{indentation}ring_ptr: {cq["ring_ptr"]}')
    entries = gdb.parse_and_eval("emper::io::IoContext::workerIo->getCqEntries()")
    print(f'{indentation}entries: {entries}')

    flags = gdb.parse_and_eval("emper::io::IoContext::workerIo->getCqFlags()")
    print(f'{indentation}flags: {flags}')

    head = gdb.parse_and_eval("emper::io::IoContext::workerIo->getCqHeadSafe()")
    print(f'{indentation}head: {head}')

    tail = gdb.parse_and_eval("emper::io::IoContext::workerIo->getCqTailSafe()")
    print(f'{indentation}tail: {tail}')

    ready = tail - head
    print(f'{indentation}ready: {ready}')

    # print(io['ring']['cq'])


def print_uring(io, indentation=''):
    uring = io['ring']
    print(f'{indentation}io_uring:')
    indentation += INDENTATION

    print(f'{indentation}ring_fd: {uring["ring_fd"]}')
    print(f'{indentation}flags: {uring["flags"]}')
    print(f'{indentation}features: {uring["features"]}')

    print_sq(io, indentation)
    print_cq(io, indentation)


def print_io(io, indentation=''):
    print(f'{indentation}IoContext:')
    indentation += INDENTATION

    print(f'{indentation}needsCqLock: {io["needsCqLock"]}')
    print(f'{indentation}locklessCq: {io["locklessCq"]}')

    print(f'{indentation}notificationEventFd: {io["notificationEventFd"]}')

    submitter = io["submitter"]
    if not submitter.address:
        print(f'{indentation}submitter: {submitter}')

    # print(f'{indentation}{io["CQE_BATCH_SIZE"]}')
    print(f'{indentation}reqs_in_uring: {io["reqs_in_uring"]["_M_i"]}')
    print(f'{indentation}preparedSqes: {io["preparedSqes"]}')
    print(f'{indentation}uringFutureSet: {io["uringFutureSet"]["_set"]}')

    print_uring(io, indentation=indentation)


def print_anywhere_queue(runtime, indentation=''):
    print(f'{indentation}AnywhereQueue:')
    indentation += INDENTATION
    scheduler = runtime["scheduler"]
    print(f'{indentation}{scheduler["scheduleAnywhereQueue"]["queue"]}')


def print_stats(runtime, indentation=''):
    print(f'{indentation}Stats:')
    indentation += INDENTATION
    print_stats_cmd = f'p ((Runtime*){runtime.address})->printStats(std::cout, false)'
    stats_str = gdb.execute(print_stats_cmd, False, True)
    for l in stats_str.splitlines():
        print(f'{indentation}{l}')


def print_bt():
    print(gdb.execute("bt", False, True))


def dump():
    printed_once = False

    threads = gdb.selected_inferior().threads()
    for thread in threads:
        # switch to thread
        thread.switch()

        # A thread may be in a frame from a non emper source file and
        # thus gdb can not use the emper namespace.
        # Our solution is to walk up the stack until we are in a emper function
        emper_found = False
        cur = gdb.selected_frame()
        while (True):
            cur_sym_tab = cur.find_sal().symtab
            if cur_sym_tab and 'emper' in cur_sym_tab.filename:
                emper_found = True
                break

            next = cur.older()
            if not next:
                break
            next.select()
            cur = next

        if not emper_found:
            print(f"Thread {thread.name} found not executing any emper code")
            print_bt()
            continue

        runtime = gdb.parse_and_eval("Runtime::currentRuntime").dereference()
        if not runtime:
            print(f"Thread {thread} found without an active runtime")
            print_bt()
            continue

        if not printed_once:
            print_anywhere_queue(runtime)
            print()
            print_stats(runtime)
            print()
            printed_once = True

        worker = gdb.parse_and_eval("Worker::currentWorker").dereference()
        if worker.address == 0:
            print(f"Non worker thread {thread.name}")
            print_bt()
            continue

        worker_id = worker['workerId']
        print(f"## Worker {worker_id} ##")
        print_bt()

        queue = gdb.parse_and_eval("AbstractWorkStealingScheduler::queue")
        print_queue(queue)
        print()

        io = gdb.parse_and_eval("emper::io::IoContext::workerIo").dereference()
        print_io(io)
        print()



if __name__ == '__main__':
    # install frame filter
    WorkerFrameFilter()
    setup()
    dump()
    cleanup()
